/*
 * NaryOperationAbstraction.hpp
 *
 *  Created on: 20. 8. 2017
 *	  Author: Jan Travnicek
 */

#ifndef _NARY_OPERATION_ABSTRACTION_HPP_
#define _NARY_OPERATION_ABSTRACTION_HPP_

#include <abstraction/ValueOperationAbstraction.hpp>
#include <alib/tuple>
#include <alib/memory>
#include <registry/Registry.h>

namespace abstraction {

template < class ReturnType, class ... ParamTypes >
class NaryOperationAbstraction : public ValueOperationAbstraction < ReturnType > {
protected:
	ext::tuple < std::shared_ptr < ValueProvider < ParamTypes > > ... > m_params;
	std::array < bool, sizeof ... ( ParamTypes ) > m_moves;

private:
	virtual bool attachInput ( const std::shared_ptr < OperationAbstraction > & input, unsigned index, bool move ) override {
		if ( index >= m_moves.size ( ) )
			throw std::invalid_argument ( "Parameter index " + ext::to_string ( index ) + " out of bounds.");

		auto attachCallback = [ & ] ( auto & param ) {
			if ( param != nullptr )
				return false;

			typename std::decay < decltype ( param )>::type validData = std::dynamic_pointer_cast < typename std::decay < decltype ( param ) >::type::element_type > ( input->getProxyAbstraction ( ) );
			if ( validData ) {
				m_moves [ index ] = move;
				param = validData;
				return true;
			} else {
				return false;
			}
		};

		return ext::call_on_nth < bool > ( m_params, index, attachCallback );
	}

	virtual bool detachInput ( unsigned index ) override {
		if ( index >= m_moves.size ( ) )
			throw std::invalid_argument ( "Parameter index " + ext::to_string ( index ) + " out of bounds.");

		m_moves [ index ] = false;

		auto detachCallback = [ & ] ( auto & param ) {
			param = nullptr;
			return true;
		};

		return ext::call_on_nth < bool > ( m_params, index, detachCallback );
	}

public:
	NaryOperationAbstraction ( ) {
		for ( unsigned i = 0; i < sizeof ... ( ParamTypes ); ++ i )
			m_moves [ i ] = false;
	}

	virtual bool inputsReady ( ) const override {
		auto readyCallback = [ ] ( const auto & param ) {
			return ( bool ) param && param->isReady ( );
		};

		return ext::all_of ( m_params, readyCallback );
	}

	virtual bool inputsAttached ( ) const override {
		auto attachedCallback = [ ] ( const auto & param ) {
			return ( bool ) param;
		};

		return ext::all_of ( m_params, attachedCallback );
	}

	virtual bool eval ( ) override {
		if ( ! inputsAttached ( ) )
			return false;

		if ( this->cached ( ) )
			return true;

		auto evalCallback = [ ] ( const auto & param ) {
			return param->eval ( );
		};

		if ( ! ext::all_of ( m_params, evalCallback ) )
			return false;

		return this->run ( );
	}

	virtual unsigned numberOfParams ( ) const override {
		return sizeof ... ( ParamTypes );
	}

	virtual ext::type_index getParamTypeIndex ( unsigned index ) const override {
		auto callback = [ & ] ( auto & param ) {
			return ext::type_index ( typeid ( typename std::decay < decltype ( param ) >::type::element_type::return_type ) );
		};
		return ext::call_on_nth < ext::type_index > ( m_params, index, callback );
	}

};

} /* namespace abstraction */

#endif /* _NARY_OPERATION_ABSTRACTION_HPP_ */
