/*
 * NFADeterminizer.cpp
 *
 *  Created on: 16. 1. 2014
 *	  Author: Jan Travnicek
 */

#include "common/RHDPDACommon.h"

#include <automaton/PDA/RealTimeHeightDeterministicNPDA.h>
#include <alib/set>
#include <global/GlobalData.h>

namespace automaton {

namespace determinize {

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
ext::set<ext::variant<EpsilonType, InputSymbolType>> getRetPartitioning(const automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & n, const ext::set < ext::pair < StateType, StateType > > & state) {
	ext::set<ext::variant<EpsilonType, InputSymbolType>> ret;

	const ext::set<StateType> dSubSet = retrieveDSubSet(state);

	for(const auto& transition : n.getReturnTransitions()) {
		if(dSubSet.count(std::get<0>(transition.first)))
			ret.insert(std::get<1>(transition.first));
	}

	return ret;
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
ext::set<ext::variant<EpsilonType, InputSymbolType>> getCallPartitioning(const automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & n, const ext::set < ext::pair < StateType, StateType > > & state) {
	ext::set<ext::variant<EpsilonType, InputSymbolType>> call;

	const ext::set<StateType> dSubSet = retrieveDSubSet(state);

	for(const auto& transition : n.getCallTransitions()) {
		if(dSubSet.count(transition.first.first))
			call.insert(transition.first.second);
	}

	return call;
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
ext::set<ext::variant<EpsilonType, InputSymbolType>> getLocalPartitioning(const automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & n, const ext::set < ext::pair < StateType, StateType > > & state) {
	ext::set<ext::variant<EpsilonType, InputSymbolType>> local;

	const ext::set<StateType> dSubSet = retrieveDSubSet(state);

	for(const auto& transition : n.getLocalTransitions()) {
		if(dSubSet.count(transition.first.first))
			local.insert(transition.first.second);
	}

	return local;
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
automaton::RealTimeHeightDeterministicDPDA < InputSymbolType, EpsilonType, ext::pair < ext::set < ext::pair < StateType, StateType > >, ext::variant < EpsilonType, InputSymbolType > >, ext::set < ext::pair < StateType, StateType > > > Determinize::determinize ( const automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & n ) {
	using DeterministicStateType = ext::set < ext::pair < StateType, StateType > >;
	using DeterministicPushdownStoreSymbolType = ext::pair < DeterministicStateType, ext::variant < EpsilonType, InputSymbolType > >;

	DeterministicStateType initialLabel = createIdentity(n.getInitialStates());
	DeterministicPushdownStoreSymbolType bottom = ext::make_pair ( DeterministicStateType { }, ext::variant < EpsilonType, InputSymbolType > ( alphabet::BottomOfTheStackSymbol::instance < InputSymbolType > ( ) ) );

	automaton::RealTimeHeightDeterministicDPDA < InputSymbolType, EpsilonType, DeterministicPushdownStoreSymbolType, DeterministicStateType > d ( initialLabel, bottom );
	d.setInputAlphabet ( n.getInputAlphabet ( ) );

	ext::set < ext::pair < DeterministicStateType, DeterministicPushdownStoreSymbolType > > rubbishReturnTransitions;

	for(;;) {
		ext::set<ext::pair<DeterministicStateType, DeterministicPushdownStoreSymbolType>> stateSymbols = existsDirtyStateSymbol(d, n);

		for ( const ext::pair < DeterministicStateType, DeterministicPushdownStoreSymbolType > & rubbishStateSymbols : rubbishReturnTransitions ) {
			stateSymbols.erase ( rubbishStateSymbols );
		}

		ext::set<DeterministicStateType> states = existsDirtyState(d, n);

		if(stateSymbols.empty() && states.empty()) break;

		for ( const ext::pair<DeterministicStateType, DeterministicPushdownStoreSymbolType> & stateSymbol : stateSymbols ) {
			if ( common::GlobalData::verbose )
				common::Streams::log << "Dirty state symbol: " << stateSymbol << std::endl;

			ext::set<ext::variant<EpsilonType, InputSymbolType>> retPart = getRetPartitioning(n, stateSymbol.first);

			const DeterministicStateType & state = stateSymbol.first;
			const DeterministicPushdownStoreSymbolType & pdaSymbol = stateSymbol.second;

			for(const ext::variant < EpsilonType, InputSymbolType > & input : retPart ) {
				DeterministicStateType to;

				if ( pdaSymbol == d.getBottomOfTheStackSymbol ( ) )
					to = retInitial ( state, input, n );
				else
					to = ret ( state, pdaSymbol, input, n );

				if ( ! to.empty ( ) ) {
					d.addState ( to );
					d.addReturnTransition ( state, input, pdaSymbol, std::move ( to ) );
				} else
					rubbishReturnTransitions.insert ( ext::make_pair ( state, pdaSymbol ) );
			}
		}

		for(const auto& state : states) {
			if ( common::GlobalData::verbose )
				common::Streams::log << "Dirty state: " << state << std::endl;

			ext::set<ext::variant<EpsilonType, InputSymbolType>> localPart = getLocalPartitioning(n, state);
			ext::set<ext::variant<EpsilonType, InputSymbolType>> callPart = getCallPartitioning(n, state);

			for ( const ext::variant < EpsilonType, InputSymbolType > & input : localPart ) {
				DeterministicStateType to = local ( state, input, n );
				if ( ! to.empty ( ) ) {
					d.addState ( to );
					d.addLocalTransition ( state, input, std::move ( to ) );
				}
			}
			for ( const ext::variant < EpsilonType, InputSymbolType > & input : callPart ) {
				DeterministicStateType to = call ( state, input, n );
				if ( ! to.empty ( ) ) {
					ext::pair < DeterministicStateType, ext::variant < EpsilonType, InputSymbolType > > pdaSymbol = ext::make_pair ( state, input );

					d.addState ( to );
					d.addPushdownStoreSymbol ( pdaSymbol );
					d.addCallTransition ( state, input, std::move ( to ), std::move ( pdaSymbol ) );
				}
			}
		}
	}

	const ext::set<StateType>& finalLabels = n.getFinalStates();
	for(const DeterministicStateType & state : d.getStates()) {
		const ext::set<StateType> labels = retrieveDSubSet(state);

		if(!ext::excludes(finalLabels.begin(), finalLabels.end(), labels.begin(), labels.end())) {
			d.addFinalState(state);
		}
	}

	if ( common::GlobalData::verbose ) {
		common::Streams::log << "Rubbish return transitions: " << rubbishReturnTransitions << std::endl;
	}

	return d;
}

} /* namespace determinize */

} /* namespace automaton */
