/*
 * ReachableStates.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Created on: 23. 3. 2014
 *	  Author: Tomas Pecka
 */

#ifndef REACHABLE_STATES_H_
#define REACHABLE_STATES_H_

#include <alib/algorithm>
#include <alib/deque>
#include <alib/set>
#include <alib/map>

#include <automaton/FSM/ExtendedNFA.h>
#include <automaton/FSM/CompactNFA.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/FSM/DFA.h>

#include <automaton/Automaton.h>

namespace automaton {

namespace properties {

/**
 * Find all reachable states of an automaton
 * A state is reachable if there exists any sequence of transitions from the initial state to it.
 */
class ReachableStates {
public:
	/**
	 * Finds all reachable states of a finite automaton.
	 * Using closure implementation of the BFS algorithm.
	 *
	 * @tparam T type of finite automaton
	 * @param fsm automaton
	 * @return set of reachable states from the initial state of @p fsm
	 */
	template < class T >
	static ext::set < typename automaton::StateTypeOfAutomaton < T > > reachableStates( const T & fsm );

	/**
	 * Finds all reachable states of a nondetermistic finite automaton with multiple initial states.
	 * Using closure implementation of the BFS algorithm.
	 *
	 * @overload
	 *
	 * @tparam SymbolType Type for the input symbols.
	 * @tparam StateType Type for the states.
	 * @param fsm nondeterministic finite automaton with multiple initial states
	 * @return set of reachable states from the initial state of @p fsm
	 */
	template < class SymbolType, class StateType >
	static ext::set<StateType> reachableStates( const automaton::MultiInitialStateNFA < SymbolType, StateType > & fsm );

	/**
	 * Finds all reachable states of a deterministic finite automaton.
	 * Using closure implementation of the BFS algorithm.
	 *
	 * @overload
	 *
	 * @tparam SymbolType Type for the input symbols.
	 * @tparam StateType Type for the states.
	 * @param fsm nondeterministic automaton with multiple initial states
	 * @return set of reachable states from the initial state of @p fsm
	 */
	template < class SymbolType, class StateType >
	static ext::set<StateType> reachableStates( const automaton::DFA < SymbolType, StateType > & fsm );
};

template < class T >
ext::set < typename automaton::StateTypeOfAutomaton < T > > ReachableStates::reachableStates( const T & fsm ) {
	using StateType = typename automaton::StateTypeOfAutomaton < T >;

	// 1a
	ext::deque<ext::set<StateType>> Qi;
	Qi.push_back( ext::set<StateType>( ) );
	Qi.at( 0 ) = {fsm.getInitialState( )};

	int i = 1;

	// 1bc
	while( true ) {
		Qi.push_back( Qi.at( i - 1 ) );

		for( const auto & p : Qi.at( i - 1 ) )
			for( const auto & transition : fsm.getTransitionsFromState( p ) )
				Qi.at( i ).insert( transition.second.begin(), transition.second.end() );

		if( Qi.at( i ) == Qi.at( i - 1 ) )
			break;

		i = i + 1;
	}

	return Qi.at( i );
}

template < class SymbolType, class StateType >
ext::set<StateType> ReachableStates::reachableStates( const automaton::MultiInitialStateNFA < SymbolType, StateType > & fsm ) {
	// 1a
	ext::deque<ext::set<StateType>> Qi;
	Qi.push_back( ext::set<StateType>( ) );
	Qi.at( 0 ) = fsm.getInitialStates( );

	int i = 1;

	// 1bc
	while( true ) {
		Qi.push_back( Qi.at( i - 1 ) );

		for( const auto & p : Qi.at( i - 1 ) )
			for( const auto & transition : fsm.getTransitionsFromState( p ) )
				Qi.at( i ).insert( transition.second.begin(), transition.second.end() );

		if( Qi.at( i ) == Qi.at( i - 1 ) )
			break;

		i = i + 1;
	}

	return Qi.at( i );
}

template < class SymbolType, class StateType >
ext::set<StateType> ReachableStates::reachableStates( const automaton::DFA < SymbolType, StateType > & fsm ) {
	// 1a
	ext::deque<ext::set<StateType>> Qi;
	Qi.push_back( ext::set<StateType>( ) );
	Qi.at( 0 ). insert( fsm.getInitialState( ) );

	int i = 1;

	// 1bc
	while( true ) {
		Qi.push_back( Qi.at( i - 1 ) );

		for( const auto & p : Qi.at( i - 1 ) )
			for( const auto & transition : fsm.getTransitionsFromState( p ) )
				Qi.at( i ).insert( transition.second );

		if( Qi.at( i ) == Qi.at( i - 1 ) )
			break;

		i = i + 1;
	}

	return Qi.at( i );
}

} /* namespace properties */

} /* namespace automaton */

#endif /* REACHABLE_STATES_H_ */
