/*
 * UnboundedRegExpElement.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef VISITOR_HPP_
#define VISITOR_HPP_

#include <alib/utility>
#include <alib/tuple>

namespace core {

/**
 * \brief
 * Class implementing an actual visitor interface to the visitor.
 *
 * The interface is represented by the return type, the visitor class implementing the callbacks, and additional params of the call.
 *
 */
template < class ReturnType, class Visitor, class ... Params >
class VisitorContextAux {
	/**
	 * \brief
	 * A place for the result.
	 */
	typename std::aligned_storage < ext::SizeOf < ReturnType >::size, ext::AlignOf < ReturnType >::align >::type result;

	/**
	 * \brief
	 * A place for the params.
	 */
	ext::tuple < Params ... > m_params;

public:
	/**
	 * \brief
	 * Constructor for initialisation of the visitor context, i.e. additional call parameters.
	 */
	VisitorContextAux ( Params && ... params ) : m_params ( std::forward < Params > ( params ) ... ) {
	}

	/**
	 * \brief
	 * Call method to the visitors visit method. The actuall class of visited object is already evaluated here.
	 *
	 * This call method is enabled if the return type is not void.
	 */
	template < class Inherit, size_t ... Indexes, class ReturnType2 = ReturnType >
	void call ( const Inherit & inherit, std::index_sequence < Indexes ... >, typename std::enable_if < !std::is_void < ReturnType2 >::value && std::is_same < ReturnType, ReturnType2 >::value >::type * = 0 ) {
		new ( & result ) ReturnType ( Visitor::visit ( inherit, std::forward < Params > ( std::get < Indexes > ( m_params ) ) ... ) );
	}


	/**
	 * \brief
	 * Call method to the visitors visit method. The actuall class of visited object is already evaluated here.
	 *
	 * This call method is enabled if the return type is void.
	 */
	template < class Inherit, size_t ... Indexes, class ReturnType2 = ReturnType >
	void call ( const Inherit & inherit, std::index_sequence < Indexes ... >, typename std::enable_if < std::is_void < ReturnType2 >::value && std::is_same < ReturnType, ReturnType2 >::value >::type * = 0 ) {
		ReturnType ( Visitor::visit ( inherit, std::forward < Params > ( std::get < Indexes > ( m_params ) ) ... ) );
	}

	/**
	 * \brief
	 * Visit result getter.
	 *
	 * Enabled if the return type is not void.
	 */
	template < class ReturnType2 = ReturnType >
	ReturnType getResult ( typename std::enable_if < !std::is_void < ReturnType2 >::value && std::is_same < ReturnType, ReturnType2 >::value >::type * = 0 ) {
		ReturnType res = reinterpret_cast < ReturnType && > ( result );

		 // reinterpret_cast < ReturnType * > ( & result )->~ReturnType ( ); // not needed to delete; data gets moved from on previous line
		return res;
	}

	/**
	 * \brief
	 * Visit result getter.
	 *
	 * Enabled if the return type not void.
	 */
	template < class ReturnType2 = ReturnType >
	void getResult ( typename std::enable_if < std::is_void < ReturnType2 >::value && std::is_same < ReturnType, ReturnType2 >::value >::type * = 0 ) {
	}

};

} /* namespace core */

#endif /* VISITOR_HPP_ */
