/*
 * XmlApi.hpp
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef XML_API_HPP_
#define XML_API_HPP_

#include <alib/functional>
#include <alib/deque>
#include <alib/map>
#include <alib/string>
#include <alib/memory>
#include <alib/typeinfo>
#include "sax/Token.h"
#include "sax/FromXMLParserHelper.h"

#include "object/Object.h"
#include "object/AnyObject.h"

#include <global/GlobalData.h>

#include "exception/CommonException.h"

namespace core {

template < typename T >
struct xmlApi;

class xmlApiInputContext : public ext::deque < sax::Token >::iterator {
	ext::map < int, object::Object > idToInstanceContexts;

public:
	xmlApiInputContext ( ext::deque < sax::Token >::iterator iter ) : ext::deque < sax::Token >::iterator ( iter ) {
	}

	ext::map < int, object::Object > & idToInstance ( ) {
		return idToInstanceContexts;
	}

};

class xmlApiOutputContext : public ext::deque < sax::Token > {
	ext::map < object::Object, int > instanceToIdContexts;
	int idMaxContext;

public:
	xmlApiOutputContext ( ) : idMaxContext ( 0 ) {
	}

	ext::map < object::Object, int > & instanceToId ( ) {
		return instanceToIdContexts;
	}

	int & idMax ( ) {
		return idMaxContext;
	}

};

template < >
struct xmlApi < object::Object > {
private:
	class GroupParser {
	public:
		virtual object::Object parse ( ext::deque < sax::Token >::iterator & input ) = 0;

		virtual ~GroupParser ( ) noexcept = default;
	};

	static ext::map < std::string, std::unique_ptr < GroupParser > > & parseFunctions ( ) {
		static ext::map < std::string, std::unique_ptr < GroupParser > > res;

		return res;
	}

	template < class Type >
	class ParserRegister : public GroupParser {
	public:
		virtual object::Object parse ( ext::deque < sax::Token >::iterator & input ) {
			return object::Object ( xmlApi < Type >::parse ( input ) );
		}

	};

	class GroupComposer {
	public:
		virtual void compose ( ext::deque < sax::Token > & input, const object::Object & group ) = 0;

		virtual ~GroupComposer( ) noexcept = default;
	};

	static ext::map < std::string, std::unique_ptr < GroupComposer > > & composeFunctions ( ) {
		static ext::map < std::string, std::unique_ptr < GroupComposer > > res;

		return res;
	}

	template < class Type >
	class ComposerRegister : public GroupComposer {
	public:
		virtual void compose ( ext::deque < sax::Token > & input, const object::Object & group ) {
			xmlApi < Type >::compose ( input, static_cast < const object::AnyObject < Type > & > ( group.getData ( ) ).getData ( ) );
		}

	};

public:
	template < class Type >
	static void registerXmlReader ( ) {
		bool res = parseFunctions ( ).insert ( std::make_pair ( xmlApi < Type >::xmlTagName(), std::unique_ptr < GroupParser > ( new ParserRegister < Type > ( ) ) ) ).second;
		if ( ! res ) {
			std::string groupName = ext::to_string < object::Object > ( );
			std::string typeName = ext::to_string < Type > ( );

			throw::exception::CommonException ( "Parse callback of " + typeName + " already registered in group " + groupName + "." );
		}
	}

	template < class Type >
	static void registerXmlWriter ( ) {
		bool res = composeFunctions ( ).insert ( std::make_pair ( ext::to_string < object::AnyObject < Type > > ( ), std::unique_ptr < GroupComposer > ( new ComposerRegister < Type > ( ) ) ) ).second;
		if ( ! res ) {
			std::string groupName = ext::to_string < object::Object > ( );
			std::string typeName = ext::to_string < Type > ( );

			throw::exception::CommonException ( "Parse callback of " + typeName + " already registered in group " + groupName + "." );
		}
	}

	static object::Object parseRef ( xmlApiInputContext & input ) {
		sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "Ref" );
		sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ATTRIBUTE, "id" );
		int id = ext::from_string < int > ( sax::FromXMLParserHelper::popTokenData ( input, sax::Token::TokenType::CHARACTER ) );
		sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ATTRIBUTE, "id" );
		sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "Ref" );
		ext::map < int, object::Object >::iterator elem = input.idToInstance ( ).find ( id );

		if ( elem == input.idToInstance ( ).end ( ) ) {
			common::Streams::err << input.idToInstance ( ) << std::endl;
			throw exception::CommonException ( "XML Inconsistent ( id not found " + ext::to_string ( id ) + " )" );
		}

		return elem->second;
	}

	static object::Object parseUnique ( xmlApiInputContext & input ) {
		sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "UniqueObject" );
		object::Object baseObject ( parse ( input ) );
		sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "id" );
		unsigned id = ext::from_string < unsigned > ( sax::FromXMLParserHelper::popTokenData ( input, sax::Token::TokenType::CHARACTER ) );
		sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "id" );
		sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "UniqueObject" );
		baseObject += id;
		return baseObject;
	}

	static object::Object parseObject ( xmlApiInputContext & input, const std::string & tagName ) {
		typename ext::map < std::string, std::unique_ptr < GroupParser > >::iterator callback = parseFunctions ( ).find ( tagName );

		if ( callback == parseFunctions ( ).end ( ) ) throw exception::CommonException ( "Parse callback for " + tagName + " tag not registered." );

		return callback->second->parse ( input );
	}

	static object::Object parse ( xmlApiInputContext & input, const std::string & tagName ) {
		if ( /* common::GlobalData::optimizeXml && */ tagName == "Ref" ) {
			return parseRef ( input );
		} else if ( tagName == "UniqueObject" ) {
			return parseUnique ( input );
		} else {
			return parseObject ( input, tagName );
		}
	}

	static object::Object parse ( ext::deque < sax::Token >::iterator & data ) {
		xmlApiInputContext & input = ( xmlApiInputContext & ) data;
		sax::FromXMLParserHelper::skipAttributes ( input, sax::Token::TokenType::START_ELEMENT );

		const std::string & tagName = sax::FromXMLParserHelper::getTokenData ( input, sax::Token::TokenType::START_ELEMENT );

		/* find out if this is a base for reference */
		bool ref = sax::FromXMLParserHelper::isToken ( input + 1, sax::Token::TokenType::START_ATTRIBUTE, "ref" );
		int id = 0;
		if ( ref ) {
			ext::deque < sax::Token >::iterator tmp = input + 2;
			id = ext::from_string < unsigned > ( sax::FromXMLParserHelper::popTokenData ( tmp, sax::Token::TokenType::CHARACTER ) );
		}

		object::Object res = parse ( input, tagName );

		/* if object is a base of reference, register it */
		if ( ref )
			input.idToInstance ( ).insert ( std::make_pair ( id, res ) );

		return res;
	}

	static bool first ( const ext::deque < sax::Token >::const_iterator & input ) {
		const std::string & tagName = sax::FromXMLParserHelper::getTokenData ( input, sax::Token::TokenType::START_ELEMENT );

		if ( tagName == "Ref" ) return true;

		typename ext::map < std::string, std::unique_ptr < GroupParser > >::iterator callback = parseFunctions ( ).find ( tagName );

		if ( callback == parseFunctions ( ).end ( ) ) return false;

		return true;
	}

	static std::string xmlTagName ( ) {
		std::string target = ext::to_string < object::Object > ( );

		throw ::exception::CommonException ( "Type " + target + " does not have xmlTagName." );
	}

	static void composeRef ( xmlApiOutputContext & output, typename ext::map < object::Object, int >::iterator & elem ) {
		output.emplace_back ( "Ref", sax::Token::TokenType::START_ELEMENT );
		output.emplace_back ( "id", sax::Token::TokenType::START_ATTRIBUTE );
		output.emplace_back ( ext::to_string ( elem->second ), sax::Token::TokenType::CHARACTER );
		output.emplace_back ( "id", sax::Token::TokenType::END_ATTRIBUTE );
		output.emplace_back ( "Ref", sax::Token::TokenType::END_ELEMENT );
	}

	static void composeUnique ( xmlApiOutputContext & output, const object::Object & data ) {
		output.emplace_back ( "UniqueObject", sax::Token::TokenType::START_ELEMENT );
		composeObject ( output, data );
		output.emplace_back ( "id", sax::Token::TokenType::START_ELEMENT );
		output.emplace_back ( ext::to_string ( data.getId ( ) ), sax::Token::TokenType::CHARACTER );
		output.emplace_back ( "id", sax::Token::TokenType::END_ELEMENT );
		output.emplace_back ( "UniqueObject", sax::Token::TokenType::END_ELEMENT );
	}

	static void composeObject ( xmlApiOutputContext & output, const object::Object & data ) {
		const auto & content = data.getData ( );
		std::string type = ext::to_string ( ext::type_index ( typeid ( content ) ) );
		typename ext::map < std::string, std::unique_ptr < GroupComposer > >::iterator callback = composeFunctions ( ).find ( type );

		if ( callback == composeFunctions ( ).end ( ) ) throw exception::CommonException ( "Compose callback for " + type + " tag not registered." );

		/* encode referenced object */
		callback->second->compose ( output, data );
	}

	static void compose ( ext::deque < sax::Token > & data, const object::Object & object ) {
		xmlApiOutputContext & output = ( xmlApiOutputContext & ) data;

		typename ext::map < object::Object, int >::iterator elem = output.instanceToId ( ).find ( object );

		/* get current position in token stream */
		size_t pos = output.size();
		int id = 0;
		if ( common::GlobalData::optimizeXml && elem == output.instanceToId ( ).end ( ) ) {
			id = output.idMax ( )++;
		}

		if ( common::GlobalData::optimizeXml && elem != output.instanceToId ( ).end ( ) ) {
			composeRef ( output, elem );
		} else if ( object.getId ( ) != 0 ) {
			composeUnique ( output, object );
		} else {
			composeObject ( output, object );
		}

		/* emplace referenced attribute */
		if ( common::GlobalData::optimizeXml && elem == output.instanceToId ( ).end ( ) ) {
			output.emplace ( output.begin ( ) + pos + 1, "ref", sax::Token::TokenType::START_ATTRIBUTE );
			output.emplace ( output.begin ( ) + pos + 2, ext::to_string ( id ), sax::Token::TokenType::CHARACTER );
			output.emplace ( output.begin ( ) + pos + 3, "ref", sax::Token::TokenType::END_ATTRIBUTE );

			output.instanceToId ( ).insert ( std::make_pair ( object, id ) );
		}
	}

};

} /* namespace core */

#endif /* XML_API_HPP_ */
