/*
 * functional.hpp
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef _FUNCTIONAL_HPP_
#define _FUNCTIONAL_HPP_

#include <functional>

#include "type_traits.hpp"
#include "typeindex.h"

namespace ext {

template < typename _Tp = void >
struct less;

template < typename _Tp >
struct less : public std::binary_function < _Tp, _Tp, bool > {
	bool operator ( ) ( const _Tp & __x, const _Tp & __y ) const {
		return __x < __y;
	}
};

template < >
struct less < void > {
	template < typename _Tp, typename _Up, typename std::enable_if < supports < std::less < > ( _Tp, _Up ) >::value >::type * = nullptr >
	auto operator ( ) ( _Tp && __t, _Up && __u ) const noexcept ( noexcept ( std::forward < _Tp > ( __t ) < std::forward < _Up > ( __u ) ) ) -> decltype ( std::forward < _Tp > ( __t ) < std::forward < _Up > ( __u ) ) {
		return std::forward < _Tp > ( __t ) < std::forward < _Up > ( __u );
	}

	template < typename _Tp, typename _Up, typename std::enable_if < ! supports < std::less < > ( _Tp, _Up ) >::value >::type * = nullptr >
	bool operator ( ) ( _Tp &&, _Up && ) const noexcept {
		return ext::type_index ( typeid ( _Tp ) ) < ext::type_index ( typeid ( _Up ) );
	}

	typedef std::__is_transparent is_transparent;
};

/**
 * \brief
 * Class extending the reference wrapper class from the standard library. Original reason is to allow its use with standard stream aggregation class.
 *
 * The class mimics the behavior of the reference wrapper from the standatd library.
 *
 * \tparam T the type of the reference inside the reference wrapper
 */
template < class T >
class reference_wrapper : public std::reference_wrapper < T > {
public:
	/**
	 * Inherit constructors of the standard reference_wrapper.
	 */
	using std::reference_wrapper < T >::reference_wrapper;

	/**
	 * Inherit operator = of the standard reference_wrapper.
	 */
	using std::reference_wrapper < T >::operator =;

#ifndef __clang__

	/**
	 * Default constructor needed by g++ since it is not inherited.
	 */
	reference_wrapper ( ) = default;

	/**
	 * Copy constructor needed by g++ since it is not inherited.
	 */
	reference_wrapper ( const reference_wrapper & other ) = default;

	/**
	 * Move constructor needed by g++ since it is not inherited.
	 */
	reference_wrapper ( reference_wrapper && other ) = default;

	/**
	 * Copy operator = needed by g++ since it is not inherited.
	 */
	reference_wrapper & operator = ( reference_wrapper && other ) = default;

	/**
	 * Move operator = needed by g++ since it is not inherited.
	 */
	reference_wrapper & operator = ( const reference_wrapper & other ) = default;
#endif

	/**
	 * Inherit the accessor of the internal reference.
	 */
	using std::reference_wrapper < T >::operator T &;
};

} /* namespace ext */

#endif /* _FUNCTIONAL_HPP_ */
