/*
 * FormalRTEAlternation.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Created on: Feb 20, 2016
 *      Author: Tomas Pecka
 */

#ifndef FORMAL_RTE_ALTERNATION_H_
#define FORMAL_RTE_ALTERNATION_H_

#include <sstream>
#include <alib/utility>

#include "FormalRTEElement.h"

#include <exception/CommonException.h>

namespace rte {

/**
 * \brief Represents the alternation operator in the regular tree expression. The node must have exactly two children.
 *
 * The structure is derived from BinaryNode that provides the children accessors.
 *
 * The node can be visited by the FormalRTEElement < SymbolType, RankType >::Visitor
 *
 * \tparam SymbolType used for the symbol part of the ranked symbol
 * \tparam RankType used for the rank part of the ranked symbol
 */
template < class SymbolType, class RankType >
class FormalRTEAlternation : public FormalRTEElement < SymbolType, RankType >, public ext::BinaryNode < FormalRTEElement < SymbolType, RankType >, FormalRTEAlternation < SymbolType, RankType > > {
	/**
	 * @copydoc regexp::FormalRTEElement < SymbolType, RankType >::accept ( ) const
	 */
	virtual void accept ( typename FormalRTEElement < SymbolType, RankType >::Visitor & visitor ) const override {
		visitor.visit ( * this );
	}

public:
	/**
	 * \brief Creates a new instance of the alternation node with explicit alternated elements
	 *
	 * \param left the first alternated element
	 * \param right the second alternated element
	 */
	explicit FormalRTEAlternation ( FormalRTEElement < SymbolType, RankType > && left, FormalRTEElement < SymbolType, RankType > && right );

	/**
	 * \brief Creates a new instance of the alternation node with explicit alternated elements
	 *
	 * \param left the first alternated element
	 * \param right the second alternated element
	 */
	explicit FormalRTEAlternation ( const FormalRTEElement < SymbolType, RankType > & left, const FormalRTEElement < SymbolType, RankType > & right );

	/**
	 * @copydoc FormalRTEElement::clone ( ) const &
	 */
	virtual FormalRTEAlternation < SymbolType, RankType > * clone ( ) const & override;

	/**
	 * @copydoc FormalRTEElement::clone ( ) &&
	 */
	virtual FormalRTEAlternation < SymbolType, RankType > * clone ( ) && override;

	/**
	 * @copydoc FormalRTEElement::testSymbol ( const common::ranked_symbol < SymbolType, RankType > & ) const
	 */
	virtual bool testSymbol ( const common::ranked_symbol < SymbolType, RankType > & symbol ) const override;

	/**
	 * @copydoc FormalRTEElement::computeMinimalAlphabet ( ext::set < common::ranked_symbol < SymbolType, RankType > > &, ext::set < common::ranked_symbol < SymbolType, RankType > > & ) const
	 */
	virtual void computeMinimalAlphabet ( ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetF, ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetK ) const override;

	/**
	 * @copydoc FormalRTEElement::checkAlphabet ( const ext::set < common::ranked_symbol < SymbolType, RankType > > &, const ext::set < common::ranked_symbol < SymbolType, RankType > > & ) const
	 */
	virtual bool checkAlphabet ( const ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetF, const ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetK ) const override;

	/**
	 * Getter of the left alternated element
	 *
	 * \return left alternated element
	 */
	const FormalRTEElement < SymbolType, RankType > & getLeftElement ( ) const;

	/**
	 * Getter of the right alternated element
	 *
	 * \return right alternated element
	 */
	const FormalRTEElement < SymbolType, RankType > & getRightElement ( ) const;

	/**
	 * Getter of the left alternated element
	 *
	 * \return left alternated element
	 */
	FormalRTEElement < SymbolType, RankType > & getLeftElement ( );

	/**
	 * Getter of the right alternated element
	 *
	 * \return right alternated element
	 */
	FormalRTEElement < SymbolType, RankType > & getRightElement ( );

	/**
	 * Setter of the left alternated element
	 *
	 * \param element left alternated element
	 */
	void setLeftElement ( FormalRTEElement < SymbolType, RankType > && element );

	/**
	 * Setter of the left alternated element
	 *
	 * \param element left alternated element
	 */
	void setLeftElement ( const FormalRTEElement < SymbolType, RankType > & element );

	/**
	 * Setter of the right alternated element
	 *
	 * \param element right alternated element
	 */
	void setRightElement ( FormalRTEElement < SymbolType, RankType > && element );

	/**
	 * Setter of the right alternated element
	 *
	 * \param element right alternated element
	 */
	void setRightElement ( const FormalRTEElement < SymbolType, RankType > & element );

	/**
	 * @copydoc base::CommonBase < FormalRTEElement < SymbolType, RankType > >::compare ( const FormalRTEElement < SymbolType, RankType > & ) const
	 */
	virtual int compare ( const FormalRTEElement < SymbolType, RankType > & other ) const override {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	/**
	 * The actual compare method
	 *
	 * \param other the other instance
	 *
	 * \returns the actual relation between two by type same node instances
	 */
	int compare ( const FormalRTEAlternation & ) const;

	/**
	 * @copydoc base::CommonBase < FormalRTEElement < SymbolType, RankType > >::operator >> ( std::ostream & ) const
	 */
	virtual void operator >>( std::ostream & out ) const override;

	/**
	 * @copydoc base::CommonBase < FormalRTEElement < SymbolType, RankType > >::operator std::string ( ) const
	 */
	virtual explicit operator std::string ( ) const override;

	/**
	 * @copydoc FormalRTEElement < SymbolType, RankType >::normalize ( ) &&
	 */
	virtual ext::smart_ptr < FormalRTEElement < DefaultSymbolType, DefaultRankType > > normalize ( ) && override {
		return ext::smart_ptr < FormalRTEElement < DefaultSymbolType, DefaultRankType > > ( new FormalRTEAlternation < DefaultSymbolType, DefaultRankType > ( std::move ( * std::move ( getLeftElement ( ) ).normalize ( ) ), std::move ( * std::move ( getRightElement ( ) ).normalize ( ) ) ) );
	}
};

template < class SymbolType, class RankType >
FormalRTEAlternation < SymbolType, RankType >::FormalRTEAlternation ( FormalRTEElement < SymbolType, RankType > && left, FormalRTEElement < SymbolType, RankType > && right ) : ext::BinaryNode < FormalRTEElement < SymbolType, RankType >, FormalRTEAlternation < SymbolType, RankType > > ( std::move ( left ), std::move ( right ) ) {
}

template < class SymbolType, class RankType >
FormalRTEAlternation < SymbolType, RankType >::FormalRTEAlternation ( const FormalRTEElement < SymbolType, RankType > & left, const FormalRTEElement < SymbolType, RankType > & right ) : FormalRTEAlternation ( ext::move_copy ( left ), ext::move_copy ( right ) ) {
}

template < class SymbolType, class RankType >
const FormalRTEElement < SymbolType, RankType > & FormalRTEAlternation < SymbolType, RankType >::getLeftElement ( ) const {
	return this->getLeft ( );
}

template < class SymbolType, class RankType >
const FormalRTEElement < SymbolType, RankType > & FormalRTEAlternation < SymbolType, RankType >::getRightElement ( ) const {
	return this->getRight ( );
}

template < class SymbolType, class RankType >
FormalRTEElement < SymbolType, RankType > & FormalRTEAlternation < SymbolType, RankType >::getLeftElement ( ) {
	return this->getLeft ( );
}

template < class SymbolType, class RankType >
FormalRTEElement < SymbolType, RankType > & FormalRTEAlternation < SymbolType, RankType >::getRightElement ( ) {
	return this->getRight ( );
}

template < class SymbolType, class RankType >
void FormalRTEAlternation < SymbolType, RankType >::setLeftElement ( const FormalRTEElement < SymbolType, RankType > & element ) {
	setLeftElement ( ext::move_copy ( element ) );
}

template < class SymbolType, class RankType >
void FormalRTEAlternation < SymbolType, RankType >::setLeftElement ( FormalRTEElement < SymbolType, RankType > && element ) {
	setLeft ( std::move ( element ) );
}

template < class SymbolType, class RankType >
void FormalRTEAlternation < SymbolType, RankType >::setRightElement ( const FormalRTEElement < SymbolType, RankType > & element ) {
	setRightElement ( ext::move_copy ( element ) );
}

template < class SymbolType, class RankType >
void FormalRTEAlternation < SymbolType, RankType >::setRightElement ( FormalRTEElement < SymbolType, RankType > && element ) {
	setRight ( std::move ( element ) );
}

template < class SymbolType, class RankType >
FormalRTEAlternation < SymbolType, RankType > * FormalRTEAlternation < SymbolType, RankType >::clone ( ) const & {
	return new FormalRTEAlternation ( * this );
}

template < class SymbolType, class RankType >
FormalRTEAlternation < SymbolType, RankType > * FormalRTEAlternation < SymbolType, RankType >::clone ( ) && {
	return new FormalRTEAlternation ( std::move ( * this ) );
}

template < class SymbolType, class RankType >
int FormalRTEAlternation < SymbolType, RankType >::compare ( const FormalRTEAlternation & other ) const {
	int res = getLeftElement ( ).compare ( other.getLeftElement ( ) );

	if ( res == 0 ) res = getRightElement ( ).compare ( other.getRightElement ( ) );

	return res;
}

template < class SymbolType, class RankType >
void FormalRTEAlternation < SymbolType, RankType >::operator >>( std::ostream & out ) const {
	out << "(FormalRTEAlternation";
	out << " " << getLeftElement ( );
	out << " " << getRightElement ( );
	out << ")";
}

template < class SymbolType, class RankType >
bool FormalRTEAlternation < SymbolType, RankType >::testSymbol ( const common::ranked_symbol < SymbolType, RankType > & symbol ) const {
	if ( getLeftElement ( ).testSymbol ( symbol ) ) return true;

	if ( getRightElement ( ).testSymbol ( symbol ) ) return true;

	return false;
}

template < class SymbolType, class RankType >
void FormalRTEAlternation < SymbolType, RankType >::computeMinimalAlphabet ( ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetF, ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetK ) const {
	getLeftElement ( ).computeMinimalAlphabet ( alphabetF, alphabetK );
	getRightElement ( ).computeMinimalAlphabet ( alphabetF, alphabetK );
}

template < class SymbolType, class RankType >
bool FormalRTEAlternation < SymbolType, RankType >::checkAlphabet ( const ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetF, const ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetK ) const {
	return getLeftElement ( ).checkAlphabet ( alphabetF, alphabetK ) && getRightElement ( ).checkAlphabet ( alphabetF, alphabetK );
}

template < class SymbolType, class RankType >
FormalRTEAlternation < SymbolType, RankType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

} /* namespace rte */

#endif /* FORMAL_RTE_ALTERNATION_H_ */
