/*
 * FormalRTEEmpty.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Created on: Feb 20, 2016
 *      Author: Tomas Pecka
 */

#ifndef FORMAL_RTE_EMPTY_H_
#define FORMAL_RTE_EMPTY_H_

#include "FormalRTEElement.h"
#include <sstream>

namespace rte {

/**
 * \brief Represents the empty expression in the regular tree expression. The node can't have any children.
 *
 * The structure is derived from NullaryNode disallowing adding any child.
 *
 * The node can be visited by the FormalRTEElement < SymbolType, RankType >::Visitor
 *
 * \tparam SymbolType used for the symbol part of the ranked symbol
 * \tparam RankType used for the rank part of the ranked symbol
 */
template < class SymbolType, class RankType >
class FormalRTEEmpty : public FormalRTEElement < SymbolType, RankType > {
	/**
	 * @copydoc regexp::FormalRTEElement < SymbolType, RankType >::accept ( ) const
	 */
	virtual void accept ( typename FormalRTEElement < SymbolType, RankType >::Visitor & visitor ) const override {
		visitor.visit ( * this );
	}

public:
	/**
	 * \brief Creates a new instance of the empty node.
	 */
	explicit FormalRTEEmpty ( );

	/**
	 * @copydoc FormalRTEElement < SymbolType, RankType >::clone ( ) const &
	 */
	virtual FormalRTEEmpty < SymbolType, RankType > * clone ( ) const & override;

	/**
	 * @copydoc FormalRTEElement < SymbolType, RankType >::clone ( ) &&
	 */
	virtual FormalRTEEmpty < SymbolType, RankType > * clone ( ) && override;

	/**
	 * @copydoc FormalRTEElement < SymbolType, RankType >::testSymbol ( const common::ranked_symbol < SymbolType, RankType > & ) const
	 */
	virtual bool testSymbol ( const common::ranked_symbol < SymbolType, RankType > & symbol ) const override;

	/**
	 * @copydoc FormalRTEElement < SymbolType, RankType >::computeMinimalAlphabet ( ext::set < common::ranked_symbol < SymbolType, RankType > > &, ext::set < common::ranked_symbol < SymbolType, RankType > > & ) const
	 */
	virtual void computeMinimalAlphabet ( ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetF, ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetK ) const override;

	/**
	 * @copydoc FormalRTEElement < SymbolType, RankType >::checkAlphabet ( const ext::set < common::ranked_symbol < SymbolType, RankType > > &, const ext::set < common::ranked_symbol < SymbolType, RankType > > & ) const
	 */
	virtual bool checkAlphabet ( const ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetF, const ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetK ) const override;

	/**
	 * @copydoc base::CommonBase < FormalRTEElement < SymbolType, RankType > >::compare ( const FormalRTEElement < SymbolType, RankType > & ) const
	 */
	virtual int compare ( const FormalRTEElement < SymbolType, RankType > & other ) const override {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	/**
	 * The actual compare method
	 *
	 * \param other the other instance
	 *
	 * \returns the actual relation between two by type same node instances
	 */
	int compare ( const FormalRTEEmpty & ) const;

	/**
	 * @copydoc base::CommonBase < FormalRTEElement < SymbolType, RankType > >::operator >> ( std::ostream & ) const
	 */
	virtual void operator >>( std::ostream & out ) const override;

	/**
	 * @copydoc base::CommonBase < FormalRTEElement < SymbolType, RankType > >::operator std::string ( ) const
	 */
	virtual explicit operator std::string ( ) const override;

	/**
	 * @copydoc FormalRTEElement::normalize ( ) &&
	 */
	virtual ext::smart_ptr < FormalRTEElement < DefaultSymbolType, DefaultRankType > > normalize ( ) && override {
		return ext::smart_ptr < FormalRTEElement < DefaultSymbolType, DefaultRankType > > ( new FormalRTEEmpty < DefaultSymbolType, DefaultRankType > ( ) );
	}
};

template < class SymbolType, class RankType >
FormalRTEEmpty < SymbolType, RankType >::FormalRTEEmpty ( ) {
	// so that default constructor is available
}

template < class SymbolType, class RankType >
FormalRTEEmpty < SymbolType, RankType > * FormalRTEEmpty < SymbolType, RankType >::clone ( ) const & {
	return new FormalRTEEmpty ( * this );
}

template < class SymbolType, class RankType >
FormalRTEEmpty < SymbolType, RankType > * FormalRTEEmpty < SymbolType, RankType >::clone ( ) && {
	return new FormalRTEEmpty ( std::move ( * this ) );
}

template < class SymbolType, class RankType >
int FormalRTEEmpty < SymbolType, RankType >::compare ( const FormalRTEEmpty & ) const {
	return 0;
}

template < class SymbolType, class RankType >
void FormalRTEEmpty < SymbolType, RankType >::operator >>( std::ostream & out ) const {
	out << "(FormalRTEEmpty)";
}

template < class SymbolType, class RankType >
bool FormalRTEEmpty < SymbolType, RankType >::testSymbol ( const common::ranked_symbol < SymbolType, RankType > & ) const {
	return false;
}

template < class SymbolType, class RankType >
void FormalRTEEmpty < SymbolType, RankType >::computeMinimalAlphabet ( ext::set < common::ranked_symbol < SymbolType, RankType > > & /*alphabetF*/, ext::set < common::ranked_symbol < SymbolType, RankType > > & /*alphabetK*/ ) const {
}

template < class SymbolType, class RankType >
bool FormalRTEEmpty < SymbolType, RankType >::checkAlphabet ( const ext::set < common::ranked_symbol < SymbolType, RankType > > & /* alphabetF */, const ext::set < common::ranked_symbol < SymbolType, RankType > > & /* alphabetK */ ) const {
	return true;
}

template < class SymbolType, class RankType >
FormalRTEEmpty < SymbolType, RankType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

} /* namespace rte */

#endif /* FORMAL_RTE_EMPTY_H_ */
