/*
 * FormalRTESubstitution.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Created on: Feb 20, 2016
 *      Author: Tomas Pecka
 */

#ifndef FORMAL_RTE_SUBSTITUTION_H_
#define FORMAL_RTE_SUBSTITUTION_H_

#include "FormalRTEElement.h"
#include "FormalRTESymbolSubst.h"

#include <exception/CommonException.h>
#include <sstream>
#include <alib/utility>

namespace rte {

/**
 * \brief Represents the concatenation operator in the regular tree expression. The node must have exactly two children.
 *
 * The structure is derived from BinaryNode that provides the children accessors.
 *
 * The node can be visited by the FormalRTEElement < SymbolType, RankType >::Visitor
 *
 * \tparam SymbolType used for the symbol part of the ranked symbol
 * \tparam RankType used for the rank part of the ranked symbol
 */
template < class SymbolType, class RankType >
class FormalRTESubstitution : public FormalRTEElement < SymbolType, RankType >, public ext::BinaryNode < FormalRTEElement < SymbolType, RankType >, FormalRTESubstitution < SymbolType, RankType > > {
	/**
	 * The substitution symbol of the node. The symbol will be substitued in left tree by right
	 */
	FormalRTESymbolSubst < SymbolType, RankType > m_substitutionSymbol;

	/**
	 * @copydoc regexp::FormalRTEElement < SymbolType, RankType >::accept ( ) const
	 */
	virtual void accept ( typename FormalRTEElement < SymbolType, RankType >::Visitor & visitor ) const override {
		visitor.visit ( * this );
	}

public:
	/**
	 * \brief Creates a new instance of the substitution node with explicit tree to substitute in, tree to substitue by and a symbol representing place of substitution
	 *
	 * \param left the tree to substitute in
	 * \param right the tree to substitute by
	 * \param substitutionSymbol the symbol representing the substitution place
	 */
	explicit FormalRTESubstitution ( FormalRTEElement < SymbolType, RankType > && left, FormalRTEElement < SymbolType, RankType > && right, FormalRTESymbolSubst < SymbolType, RankType > substitutionSymbol );

	/**
	 * \brief Creates a new instance of the substitution node with explicit tree to substitute in, tree to substitue by and a symbol representing place of substitution
	 *
	 * \param left the tree to substitute in
	 * \param right the tree to substitute by
	 * \param substitutionSymbol the symbol representing the substitution place
	 */
	explicit FormalRTESubstitution ( const FormalRTEElement < SymbolType, RankType > & left, const FormalRTEElement < SymbolType, RankType > & right, FormalRTESymbolSubst < SymbolType, RankType > substitutionSymbol );

	/**
	 * @copydoc FormalRTEElement < SymbolType, RankType >::clone ( ) const &
	 */
	virtual FormalRTESubstitution < SymbolType, RankType > * clone ( ) const & override;

	/**
	 * @copydoc FormalRTEElement < SymbolType, RankType >::clone ( ) &&
	 */
	virtual FormalRTESubstitution < SymbolType, RankType > * clone ( ) && override;

	/**
	 * @copydoc FormalRTEElement < SymbolType, RankType >::testSymbol ( const common::ranked_symbol < SymbolType, RankType > & ) const
	 */
	virtual bool testSymbol ( const common::ranked_symbol < SymbolType, RankType > & symbol ) const override;

	/**
	 * @copydoc FormalRTEElement < SymbolType, RankType >::computeMinimalAlphabet ( ext::set < common::ranked_symbol < SymbolType, RankType > > &, ext::set < common::ranked_symbol < SymbolType, RankType > > & ) const
	 */
	virtual void computeMinimalAlphabet ( ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetF, ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetK ) const override;

	/**
	 * @copydoc FormalRTEElement < SymbolType, RankType >::checkAlphabet ( const ext::set < common::ranked_symbol < SymbolType, RankType > > &, const ext::set < common::ranked_symbol < SymbolType, RankType > > & ) const
	 */
	virtual bool checkAlphabet ( const ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetF, const ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetK ) const override;

	/**
	 * Getter of the tree to substitute in
	 *
	 * \return tree to substitute in
	 */
	const FormalRTEElement < SymbolType, RankType > & getLeftElement ( ) const;

	/**
	 * Getter of the tree to substitute by
	 *
	 * \return tree to substitute by
	 */
	const FormalRTEElement < SymbolType, RankType > & getRightElement ( ) const;

	/**
	 * Getter of the substitution symbol
	 *
	 * \return the substitution symbol
	 */
	const FormalRTESymbolSubst < SymbolType, RankType > & getSubstitutionSymbol ( ) const;

	/**
	 * Getter of the tree to substitute in
	 *
	 * \return tree to substitute in
	 */
	FormalRTEElement < SymbolType, RankType > & getLeftElement ( );

	/**
	 * Getter of the tree to substitute by
	 *
	 * \return tree to substitute by
	 */
	FormalRTEElement < SymbolType, RankType > & getRightElement ( );

	/**
	 * Getter of the substitution symbol
	 *
	 * \return the substitution symbol
	 */
	FormalRTESymbolSubst < SymbolType, RankType > & getSubstitutionSymbol ( );

	/**
	 * Setter of the tree to substitute in
	 *
	 * \param element the tree to substitute in
	 */
	void setLeftElement ( const FormalRTEElement < SymbolType, RankType > & element );

	/**
	 * Setter of the tree to substitute by
	 *
	 * \param element the tree to substitute by
	 */
	void setLeftElement ( FormalRTEElement < SymbolType, RankType > && element );

	/**
	 * Setter of the tree to substitute in
	 *
	 * \param element the tree to substitute in
	 */
	void setRightElement ( const FormalRTEElement < SymbolType, RankType > & element );

	/**
	 * Setter of the tree to substitute by
	 *
	 * \param element the tree to substitute by
	 */
	void setRightElement ( FormalRTEElement < SymbolType, RankType > && element );

	/**
	 * Setter of the substitution symbol
	 *
	 * \param element the substitution symbol
	 */
	void setSubstitutionSymbol ( FormalRTESymbolSubst < SymbolType, RankType > element );

	/**
	 * @copydoc base::CommonBase < FormalRTEElement < SymbolType, RankType > >::compare ( const FormalRTEElement < SymbolType, RankType > & ) const
	 */
	virtual int compare ( const FormalRTEElement < SymbolType, RankType > & other ) const override {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	/**
	 * The actual compare method
	 *
	 * \param other the other instance
	 *
	 * \returns the actual relation between two by type same node instances
	 */
	int compare ( const FormalRTESubstitution & ) const;

	/**
	 * @copydoc base::CommonBase < FormalRTEElement < SymbolType, RankType > >::operator >> ( std::ostream & ) const
	 */
	virtual void operator >>( std::ostream & out ) const override;

	/**
	 * @copydoc base::CommonBase < FormalRTEElement < SymbolType, RankType > >::operator std::string ( ) const
	 */
	virtual explicit operator std::string ( ) const override;

	/**
	 * @copydoc FormalRTEElement < SymbolType, RankType >::normalize ( ) &&
	 */
	virtual ext::smart_ptr < FormalRTEElement < DefaultSymbolType, DefaultRankType > > normalize ( ) && override {
		FormalRTESymbolSubst < DefaultSymbolType, DefaultRankType > subst ( alphabet::SymbolNormalize::normalizeRankedSymbol ( std::move ( m_substitutionSymbol ).getSymbol ( ) ) );
		return ext::smart_ptr < FormalRTEElement < DefaultSymbolType, DefaultRankType > > ( new FormalRTESubstitution < DefaultSymbolType, DefaultRankType > ( std::move ( * std::move ( getLeftElement ( ) ).normalize ( ) ), std::move ( * std::move ( getRightElement ( ) ).normalize ( ) ), std::move ( subst ) ) );
	}
};

template < class SymbolType, class RankType >
FormalRTESubstitution < SymbolType, RankType >::FormalRTESubstitution ( FormalRTEElement < SymbolType, RankType > && left, FormalRTEElement < SymbolType, RankType > && right, FormalRTESymbolSubst < SymbolType, RankType > substitutionSymbol ) : ext::BinaryNode < FormalRTEElement < SymbolType, RankType >, FormalRTESubstitution < SymbolType, RankType > > ( std::move ( left ), std::move ( right ) ), m_substitutionSymbol ( std::move ( substitutionSymbol ) ) {
}

template < class SymbolType, class RankType >
FormalRTESubstitution < SymbolType, RankType >::FormalRTESubstitution ( const FormalRTEElement < SymbolType, RankType > & left, const FormalRTEElement < SymbolType, RankType > & right, FormalRTESymbolSubst < SymbolType, RankType > substitutionSymbol ) : FormalRTESubstitution ( ext::move_copy ( left ), ext::move_copy ( right ), std::move ( substitutionSymbol ) ) {
}

template < class SymbolType, class RankType >
const FormalRTEElement < SymbolType, RankType > & FormalRTESubstitution < SymbolType, RankType >::getLeftElement ( ) const {
	return this->getLeft ( );
}

template < class SymbolType, class RankType >
const FormalRTEElement < SymbolType, RankType > & FormalRTESubstitution < SymbolType, RankType >::getRightElement ( ) const {
	return this->getRight ( );
}

template < class SymbolType, class RankType >
const FormalRTESymbolSubst < SymbolType, RankType > & FormalRTESubstitution < SymbolType, RankType >::getSubstitutionSymbol ( ) const {
	return m_substitutionSymbol;
}

template < class SymbolType, class RankType >
FormalRTEElement < SymbolType, RankType > & FormalRTESubstitution < SymbolType, RankType >::getLeftElement ( ) {
	return this->getLeft ( );
}

template < class SymbolType, class RankType >
FormalRTEElement < SymbolType, RankType > & FormalRTESubstitution < SymbolType, RankType >::getRightElement ( ) {
	return this->getRight ( );
}

template < class SymbolType, class RankType >
FormalRTESymbolSubst < SymbolType, RankType > & FormalRTESubstitution < SymbolType, RankType >::getSubstitutionSymbol ( ) {
	return m_substitutionSymbol;
}

template < class SymbolType, class RankType >
void FormalRTESubstitution < SymbolType, RankType >::setLeftElement ( const FormalRTEElement < SymbolType, RankType > & element ) {
	setLeftElement ( ext::move_copy ( element ) );
}

template < class SymbolType, class RankType >
void FormalRTESubstitution < SymbolType, RankType >::setLeftElement ( FormalRTEElement < SymbolType, RankType > && element ) {
	setLeft ( std::move ( element ) );
}

template < class SymbolType, class RankType >
void FormalRTESubstitution < SymbolType, RankType >::setRightElement ( const FormalRTEElement < SymbolType, RankType > & element ) {
	setRightElement ( ext::move_copy ( element ) );
}

template < class SymbolType, class RankType >
void FormalRTESubstitution < SymbolType, RankType >::setRightElement ( FormalRTEElement < SymbolType, RankType > && element ) {
	setRight ( std::move ( element ) );
}

template < class SymbolType, class RankType >
void FormalRTESubstitution < SymbolType, RankType >::setSubstitutionSymbol ( FormalRTESymbolSubst < SymbolType, RankType > symbol ) {
	m_substitutionSymbol = std::move ( symbol );
}

template < class SymbolType, class RankType >
FormalRTESubstitution < SymbolType, RankType > * FormalRTESubstitution < SymbolType, RankType >::clone ( ) const & {
	return new FormalRTESubstitution ( * this );
}

template < class SymbolType, class RankType >
FormalRTESubstitution < SymbolType, RankType > * FormalRTESubstitution < SymbolType, RankType >::clone ( ) && {
	return new FormalRTESubstitution ( std::move ( * this ) );
}

template < class SymbolType, class RankType >
int FormalRTESubstitution < SymbolType, RankType >::compare ( const FormalRTESubstitution < SymbolType, RankType > & other ) const {
	auto first = ext::tie ( getLeftElement ( ), getRightElement ( ), getSubstitutionSymbol ( ) );
	auto second = ext::tie ( other.getLeftElement ( ), other.getRightElement ( ), other.getSubstitutionSymbol ( ) );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType, class RankType >
void FormalRTESubstitution < SymbolType, RankType >::operator >>( std::ostream & out ) const {
	out << "(FormalRTESubstitution";
	out << " " << getLeftElement ( );
	out << " " << m_substitutionSymbol;
	out << " " << getRightElement ( );
	out << ")";
}

template < class SymbolType, class RankType >
bool FormalRTESubstitution < SymbolType, RankType >::testSymbol ( const common::ranked_symbol < SymbolType, RankType > & symbol ) const {
	if ( getLeftElement ( ).testSymbol ( symbol ) ) return true;

	if ( getRightElement ( ).testSymbol ( symbol ) ) return true;

	if ( m_substitutionSymbol.getSymbol ( ) == symbol ) return true;

	return false;
}

template < class SymbolType, class RankType >
void FormalRTESubstitution < SymbolType, RankType >::computeMinimalAlphabet ( ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetF, ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetK ) const {
	alphabetK.insert ( m_substitutionSymbol.getSymbol ( ) );
	getLeftElement ( ).computeMinimalAlphabet ( alphabetF, alphabetK );
	getRightElement ( ).computeMinimalAlphabet ( alphabetF, alphabetK );
}

template < class SymbolType, class RankType >
bool FormalRTESubstitution < SymbolType, RankType >::checkAlphabet ( const ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetF, const ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetK ) const {
	return alphabetK.count ( getSubstitutionSymbol ( ).getSymbol ( ) ) > 0 && getLeftElement ( ).checkAlphabet ( alphabetF, alphabetK ) && getRightElement ( ).checkAlphabet ( alphabetF, alphabetK );
}

template < class SymbolType, class RankType >
FormalRTESubstitution < SymbolType, RankType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

} /* namespace rte */

#endif /* FORMAL_RTE_SUBSTITUTION_H_ */
