/*
 * FormalRTESymbolAlphabet.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Created on: Feb 20, 2016
 *      Author: Tomas Pecka
 */

#ifndef FORMAL_RTE_SYMBOL_ALPHABET_H_
#define FORMAL_RTE_SYMBOL_ALPHABET_H_

#include "FormalRTESymbol.h"

#include <exception/CommonException.h>
#include <sstream>
#include <alib/utility>
#include <alphabet/common/SymbolNormalize.h>

namespace rte {

/**
 * \brief Represents the terminal symbol in the regular tree expression. The number of children must be the same as the arity of the symbol of the node.
 *
 * The structure is derived from VararyNode, which however does not check the number of children of the node.
 *
 * The node can be visited by the FormalRTEElement < SymbolType, RankType >::Visitor
 *
 * \tparam SymbolType used for the symbol part of the ranked symbol
 * \tparam RankType used for the rank part of the ranked symbol
 */
template < class SymbolType, class RankType >
class FormalRTESymbolAlphabet : public FormalRTESymbol < SymbolType, RankType >, public ext::VararyNode < FormalRTEElement < SymbolType, RankType >, FormalRTESymbolAlphabet < SymbolType, RankType > > {
	/**
	 * @copydoc regexp::FormalRTEElement < SymbolType, RankType >::accept ( ) const
	 */
	virtual void accept ( typename FormalRTEElement < SymbolType, RankType >::Visitor & visitor ) const override {
		visitor.visit ( * this );
	}

public:
	/**
	 * \brief Creates a new instance of the symbol node using the actual symbol to represent.
	 *
	 * \param symbol the value of the represented symbol
	 */
	explicit FormalRTESymbolAlphabet ( common::ranked_symbol < SymbolType, RankType > symbol, ext::ptr_vector < FormalRTEElement < SymbolType, RankType > > children );

	/**
	 * @copydoc FormalRTEElement < SymbolType, RankType >::clone ( ) const &
	 */
	virtual FormalRTESymbolAlphabet < SymbolType, RankType > * clone ( ) const & override;

	/**
	 * @copydoc FormalRTEElement < SymbolType, RankType >::clone ( ) &&
	 */
	virtual FormalRTESymbolAlphabet < SymbolType, RankType > * clone ( ) && override;

	/**
	 * @copydoc FormalRTEElement < SymbolType, RankType >::testSymbol ( const common::ranked_symbol < SymbolType, RankType > & ) const
	 */
	virtual bool testSymbol ( const common::ranked_symbol < SymbolType, RankType > & symbol ) const override;

	/**
	 * @copydoc FormalRTEElement < SymbolType, RankType >::computeMinimalAlphabet ( ext::set < common::ranked_symbol < SymbolType, RankType > > &, ext::set < common::ranked_symbol < SymbolType, RankType > > & ) const
	 */
	virtual void computeMinimalAlphabet ( ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetF, ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetK ) const override;

	/**
	 * @copydoc FormalRTEElement < SymbolType, RankType >::checkAlphabet ( const ext::set < common::ranked_symbol < SymbolType, RankType > > &, const ext::set < common::ranked_symbol < SymbolType, RankType > > & ) const
	 */
	virtual bool checkAlphabet ( const ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetF, const ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetK ) const override;

	/**
	 * Getter of child nodes of the rte node
	 *
	 * \return child nodes
	 */
	const ext::ptr_vector < FormalRTEElement < SymbolType, RankType > > & getElements ( ) const;

	/**
	 * Getter of child nodes of the rte node
	 *
	 * \return child nodes
	 */
	const ext::ptr_vector < FormalRTEElement < SymbolType, RankType > > & getElements ( );

	/**
	 * @copydoc base::CommonBase < FormalRTEElement < SymbolType, RankType > >::compare ( const FormalRTEElement < SymbolType, RankType > & ) const
	 */
	virtual int compare ( const FormalRTEElement < SymbolType, RankType > & other ) const override {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	/**
	 * The actual compare method
	 *
	 * \param other the other instance
	 *
	 * \returns the actual relation between two by type same node instances
	 */
	int compare ( const FormalRTESymbolAlphabet & ) const;

	/**
	 * @copydoc base::CommonBase < FormalRTEElement < SymbolType, RankType > >::operator >> ( std::ostream & ) const
	 */
	virtual void operator >>( std::ostream & out ) const override;

	/**
	 * @copydoc base::CommonBase < FormalRTEElement < SymbolType, RankType > >::operator std::string ( ) const
	 */
	virtual explicit operator std::string ( ) const override;

	/**
	 * @copydoc FormalRTEElement < SymbolType, RankType >::normalize ( ) &&
	 */
	virtual ext::smart_ptr < FormalRTEElement < DefaultSymbolType, DefaultRankType > > normalize ( ) && override {
		ext::ptr_vector < FormalRTEElement < DefaultSymbolType, DefaultRankType > > children;
		for ( FormalRTEElement < SymbolType, RankType > & element : this->getElements ( ) )
			children.push_back ( std::move ( * std::move ( element ).normalize ( ) ) );

		return ext::smart_ptr < FormalRTEElement < DefaultSymbolType, DefaultRankType > > ( new FormalRTESymbolAlphabet < DefaultSymbolType, DefaultRankType > ( alphabet::SymbolNormalize::normalizeRankedSymbol ( std::move ( this->m_symbol ) ), std::move ( children ) ) );
	}
};

template < class SymbolType, class RankType >
FormalRTESymbolAlphabet < SymbolType, RankType >::FormalRTESymbolAlphabet ( common::ranked_symbol < SymbolType, RankType > symbol, ext::ptr_vector < FormalRTEElement < SymbolType, RankType > > children ) : FormalRTESymbol < SymbolType, RankType > ( symbol ), ext::VararyNode < FormalRTEElement < SymbolType, RankType >, FormalRTESymbolAlphabet < SymbolType, RankType > > ( std::move ( children ) ) {
	if ( this->getChildren ( ).size ( ) != ( size_t ) this->getSymbol ( ).getRank ( ) )
		throw exception::CommonException ( "Symbol's rank and number of children differ. Rank is " + ext::to_string ( this->getSymbol ( ).getRank ( ) ) + ", number of children is " + ext::to_string ( this->getChildren ( ).size ( ) ) + "." );

}

template < class SymbolType, class RankType >
FormalRTESymbolAlphabet < SymbolType, RankType > * FormalRTESymbolAlphabet < SymbolType, RankType >::clone ( ) const & {
	return new FormalRTESymbolAlphabet ( * this );
}

template < class SymbolType, class RankType >
FormalRTESymbolAlphabet < SymbolType, RankType > * FormalRTESymbolAlphabet < SymbolType, RankType >::clone ( ) && {
	return new FormalRTESymbolAlphabet ( std::move ( * this ) );
}

template < class SymbolType, class RankType >
const ext::ptr_vector < FormalRTEElement < SymbolType, RankType > > & FormalRTESymbolAlphabet < SymbolType, RankType >::getElements ( ) const {
	return this->getChildren ( );
}

template < class SymbolType, class RankType >
const ext::ptr_vector < FormalRTEElement < SymbolType, RankType > > & FormalRTESymbolAlphabet < SymbolType, RankType >::getElements ( ) {
	return this->getChildren ( );
}

template < class SymbolType, class RankType >
int FormalRTESymbolAlphabet < SymbolType, RankType >::compare ( const FormalRTESymbolAlphabet < SymbolType, RankType > & other ) const {
	auto first = ext::tie ( this->getSymbol ( ), this->getChildren ( ) );
	auto second = ext::tie ( other.getSymbol ( ), other.getChildren ( ) );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType, class RankType >
void FormalRTESymbolAlphabet < SymbolType, RankType >::operator >>( std::ostream & out ) const {
	out << "(FormalRTESymbolAlphabet " << " symbol = " << this->getSymbol ( ) << " children = " << this->getChildren ( ) << "})";
}

template < class SymbolType, class RankType >
bool FormalRTESymbolAlphabet < SymbolType, RankType >::testSymbol ( const common::ranked_symbol < SymbolType, RankType > & symbol ) const {
	return symbol == this->getSymbol ( );
}

template < class SymbolType, class RankType >
void FormalRTESymbolAlphabet < SymbolType, RankType >::computeMinimalAlphabet ( ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetF, ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetK ) const {
	alphabetF.insert ( this->getSymbol ( ) );

	for ( const FormalRTEElement < SymbolType, RankType > & child : this->getElements ( ) )
		child.computeMinimalAlphabet ( alphabetF, alphabetK );
}

template < class SymbolType, class RankType >
bool FormalRTESymbolAlphabet < SymbolType, RankType >::checkAlphabet ( const ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetF, const ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetK ) const {
	return alphabetF.count ( this->getSymbol ( ) ) > 0 && std::all_of ( getElements ( ).begin ( ), getElements ( ).end ( ), [&] ( const FormalRTEElement < SymbolType, RankType > & e ) {
			return e.checkAlphabet ( alphabetF, alphabetK );
		} );
}

template < class SymbolType, class RankType >
FormalRTESymbolAlphabet < SymbolType, RankType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

} /* namespace rte */

#endif /* FORMAL_RTE_SYMBOL_ALPHABET_H_ */
