/*
 * FormalRTESymbolSubst.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Created on: Feb 20, 2016
 *      Author: Tomas Pecka
 */

#ifndef FORMAL_RTE_SUBST_SYMBOL_H_
#define FORMAL_RTE_SUBST_SYMBOL_H_

#include "FormalRTESymbol.h"

#include <exception/CommonException.h>
#include <sstream>
#include <alphabet/common/SymbolNormalize.h>

namespace rte {

/**
 * \brief Represents the substitution symbol in the regular tree expression. The node can't have any children.
 *
 * The structure is derived from NullaryNode.
 *
 * The node can be visited by the FormalRTEElement < SymbolType, RankType >::Visitor
 *
 * \tparam SymbolType used for the symbol part of the ranked symbol
 * \tparam RankType used for the rank part of the ranked symbol
 */
template < class SymbolType, class RankType >
class FormalRTESymbolSubst : public FormalRTESymbol < SymbolType, RankType >, public ext::NullaryNode < FormalRTESymbolSubst < SymbolType, RankType > > {
	/**
	 * @copydoc regexp::FormalRTEElement < SymbolType, RankType >::accept ( ) const
	 */
	virtual void accept ( typename FormalRTEElement < SymbolType, RankType >::Visitor & visitor ) const override {
		visitor.visit ( * this );
	}

public:
	/**
	 * \brief Creates a new instance of the symbol node using the actual symbol to represent.
	 *
	 * \param symbol the value of the represented symbol
	 */
	explicit FormalRTESymbolSubst ( common::ranked_symbol < SymbolType, RankType > symbol );

	/**
	 * @copydoc FormalRTEElement < SymbolType, RankType >::clone ( ) const &
	 */
	virtual FormalRTESymbolSubst < SymbolType, RankType > * clone ( ) const & override;

	/**
	 * @copydoc FormalRTEElement < SymbolType, RankType >::clone ( ) &&
	 */
	virtual FormalRTESymbolSubst < SymbolType, RankType > * clone ( ) && override;

	/**
	 * @copydoc FormalRTEElement < SymbolType, RankType >::testSymbol ( const common::ranked_symbol < SymbolType, RankType > & ) const
	 */
	virtual bool testSymbol ( const common::ranked_symbol < SymbolType, RankType > & symbol ) const override;

	/**
	 * @copydoc FormalRTEElement < SymbolType, RankType >::computeMinimalAlphabet ( ext::set < common::ranked_symbol < SymbolType, RankType > > &, ext::set < common::ranked_symbol < SymbolType, RankType > > & ) const
	 */
	virtual void computeMinimalAlphabet ( ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetF, ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetK ) const override;

	/**
	 * @copydoc FormalRTEElement < SymbolType, RankType >::checkAlphabet ( const ext::set < common::ranked_symbol < SymbolType, RankType > > &, const ext::set < common::ranked_symbol < SymbolType, RankType > > & ) const
	 */
	virtual bool checkAlphabet ( const ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetF, const ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetK ) const override;

	/**
	 * @copydoc base::CommonBase < FormalRTEElement < SymbolType, RankType > >::compare ( const FormalRTEElement < SymbolType, RankType > & ) const
	 */
	virtual int compare ( const FormalRTEElement < SymbolType, RankType > & other ) const override {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	/**
	 * The actual compare method
	 *
	 * \param other the other instance
	 *
	 * \returns the actual relation between two by type same node instances
	 */
	int compare ( const FormalRTESymbolSubst & ) const;

	/**
	 * @copydoc base::CommonBase < FormalRTEElement < SymbolType, RankType > >::operator >> ( std::ostream & ) const
	 */
	virtual void operator >>( std::ostream & out ) const override;

	/**
	 * @copydoc base::CommonBase < FormalRTEElement < SymbolType, RankType > >::operator std::string ( ) const
	 */
	virtual explicit operator std::string ( ) const override;

	/**
	 * @copydoc FormalRTEElement < SymbolType, RankType >::normalize ( ) &&
	 */
	virtual ext::smart_ptr < FormalRTEElement < DefaultSymbolType, DefaultRankType > > normalize ( ) && override {
		FormalRTESymbolSubst < DefaultSymbolType, DefaultRankType > res ( alphabet::SymbolNormalize::normalizeRankedSymbol ( std::move ( this->m_symbol ) ) );
		return ext::smart_ptr < FormalRTEElement < DefaultSymbolType, DefaultRankType > > ( std::move ( res ).clone ( ) );
	}
};

template < class SymbolType, class RankType >
FormalRTESymbolSubst < SymbolType, RankType >::FormalRTESymbolSubst ( common::ranked_symbol < SymbolType, RankType > symbol ) : FormalRTESymbol < SymbolType, RankType > ( symbol ) {
}

template < class SymbolType, class RankType >
FormalRTESymbolSubst < SymbolType, RankType > * FormalRTESymbolSubst < SymbolType, RankType >::clone ( ) const & {
	return new FormalRTESymbolSubst ( * this );
}

template < class SymbolType, class RankType >
FormalRTESymbolSubst < SymbolType, RankType > * FormalRTESymbolSubst < SymbolType, RankType >::clone ( ) && {
	return new FormalRTESymbolSubst ( std::move ( * this ) );
}

template < class SymbolType, class RankType >
int FormalRTESymbolSubst < SymbolType, RankType >::compare ( const FormalRTESymbolSubst & other ) const {
	static ext::compare < decltype ( this->getSymbol ( ) ) > comp;

	return comp ( this->getSymbol ( ), other.getSymbol ( ) );
}

template < class SymbolType, class RankType >
void FormalRTESymbolSubst < SymbolType, RankType >::operator >>( std::ostream & out ) const {
	out << "(FormalRTESymbolSubst " << " symbol = " << this->getSymbol ( ) << "})";
}

template < class SymbolType, class RankType >
bool FormalRTESymbolSubst < SymbolType, RankType >::testSymbol ( const common::ranked_symbol < SymbolType, RankType > & symbol ) const {
	return symbol == this->getSymbol ( );
}

template < class SymbolType, class RankType >
void FormalRTESymbolSubst < SymbolType, RankType >::computeMinimalAlphabet ( ext::set < common::ranked_symbol < SymbolType, RankType > > & /* alphabetF */, ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetK ) const {
	alphabetK.insert ( this->getSymbol ( ) );
}

template < class SymbolType, class RankType >
bool FormalRTESymbolSubst < SymbolType, RankType >::checkAlphabet ( const ext::set < common::ranked_symbol < SymbolType, RankType > > & /* alphabetF */, const ext::set < common::ranked_symbol < SymbolType, RankType > > & alphabetK ) const {
	return alphabetK.count ( this->getSymbol ( ) ) > 0;
}

template < class SymbolType, class RankType >
FormalRTESymbolSubst < SymbolType, RankType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

} /* namespace rte */

#endif /* FORMAL_RTE_SUBST_SYMBOL_H_ */
