/*
 * Epsilon.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef EPSILON_H_
#define EPSILON_H_

#include <alib/set>
#include <alib/vector>
#include <alib/compare>

#include <core/components.hpp>

#include <common/DefaultSymbolType.h>

#include <exception/CommonException.h>

#include <core/normalize.hpp>
#include <alphabet/common/SymbolNormalize.h>

namespace string {

class GeneralAlphabet;

/**
 * \brief
 * Representation of epsilon

 * \details
 * Definition is similar to a linear string however cyclic.
 * S = (A, C),
 * A (Alphabet) = finite set of symbols,
 * C (Content) = linear representation of cyclic string in some rotation
 *
 * Note that two same cyclic strings compare not equal in different rotation. In order to either normalize rotation or compare without influence of particular rotation, use respective algoritm.
 *
 * \tparam SymbolType used for the terminal alphabet
 */
template < class SymbolType = DefaultSymbolType >
class Epsilon final : public ext::CompareOperators < Epsilon < SymbolType > >, public core::Components < Epsilon < SymbolType >, ext::set < SymbolType >, component::Set, GeneralAlphabet > {
public:
	/**
	 * \brief Creates a new instance of the epsilon
	 */
	Epsilon ( );

	/**
	 * \brief Creates a new instance of the epsilon with a concrete alphabet.
	 *
	 * \param alphabet the initial alphabet of the string
	 */
	Epsilon ( ext::set < SymbolType > alphabet );

	/**
	 * Getter of the alphabet.
	 *
	 * \returns the alphabet of the string
	 */
	const ext::set < SymbolType > & getAlphabet ( ) const & {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	/**
	 * Getter of the alphabet.
	 *
	 * \returns the alphabet of the string
	 */
	ext::set < SymbolType > & getAlphabet ( ) && {
		return std::move ( this->template accessComponent < GeneralAlphabet > ( ).get ( ) );
	}

	/**
	 * Adder of an alphabet symbols.
	 *
	 * \param symbols the new symbols to be added to the alphabet
	 */
	void extendAlphabet ( ext::set < SymbolType > symbols ) {
		this->template accessComponent < GeneralAlphabet > ( ).add ( std::move ( symbols ) );
	}

	/**
	 * The actual compare method
	 *
	 * \param other the other instance
	 *
	 * \returns the actual relation between two by type same string instances
	 */
	int compare ( const Epsilon & other ) const;

	/**
	 * Print this object as raw representation to ostream.
	 *
	 * \param out ostream where to print
	 * \param instance object to print
	 *
	 * \returns modified output stream
	 */
	friend std::ostream & operator << ( std::ostream & out, const Epsilon & instance ) {
		return out << "(Epsilon ";
		out << "alphabet = " << instance.getAlphabet ( );
		out << ")";
	}

	/**
	 * Casts this instance to as compact as possible string representation.
	 *
	 * \returns string representation of the object
	 */
	explicit operator std::string ( ) const;

	/**
	 * Prepared defaultly constructed epsilon instance.
	 */
	static Epsilon EPSILON;
};

template < class SymbolType >
Epsilon < SymbolType >::Epsilon ( ext::set < SymbolType > alphabet ) : core::Components < Epsilon, ext::set < SymbolType >, component::Set, GeneralAlphabet > ( std::move ( alphabet ) ) {
}

template < class SymbolType >
Epsilon < SymbolType >::Epsilon() : Epsilon ( ext::set< SymbolType > ( ) ) {
}

template < class SymbolType >
int Epsilon < SymbolType >::compare(const Epsilon& other) const {
	static ext::compare<ext::set<SymbolType>> comp;
	return comp(getAlphabet(), other.getAlphabet());
}

template < class SymbolType >
Epsilon < SymbolType >::operator std::string ( ) const {
	return "E";
}

template < class SymbolType >
Epsilon < SymbolType > Epsilon < SymbolType >::EPSILON = Epsilon < SymbolType >();

} /* namespace string */

namespace core {

/**
 * Helper class specifying constraints for the epsilon's internal alphabet component.
 *
 * \tparam SymbolType used for the alphabet of the epsilon.
 */
template < class SymbolType >
class SetConstraint< string::Epsilon < SymbolType >, SymbolType, string::GeneralAlphabet > {
public:
	/**
	 * Returns true as the epsilon represents empty string.
	 *
	 * \param epsilon the tested epsilon
	 * \param symbol the tested symbol
	 *
	 * \returns true
	 */
	static bool used ( const string::Epsilon < SymbolType > &, const SymbolType & ) {
		return false;
	}

	/**
	 * Returns true as all symbols are possibly available to be in an alphabet.
	 *
	 * \param string the tested string
	 * \param symbol the tested symbol
	 *
	 * \returns true
	 */
	static bool available ( const string::Epsilon < SymbolType > &, const SymbolType & ) {
		return true;
	}

	/**
	 * All symbols are valid as symbols of an alphabet.
	 *
	 * \param string the tested string
	 * \param symbol the tested symbol
	 */
	static void valid ( const string::Epsilon < SymbolType > &, const SymbolType & ) {
	}
};

/**
 * Helper for normalisation of types specified by templates used as internal datatypes of symbols.
 *
 * \returns new instance of the string with default template parameters or unmodified instance if the template parameters were already default ones
 */
template < class SymbolType >
struct normalize < string::Epsilon < SymbolType > > {
	static string::Epsilon < > eval ( string::Epsilon < SymbolType > && value ) {
		ext::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( value ).getAlphabet ( ) );
		return string::Epsilon < > ( std::move ( alphabet ) );
	}
};

} /* namespace core */

#endif /* EPSILON_H_ */
